MODULE Comm_tools


!***************************************************************
!
!	Module for reading and writing data for analysis routines
!
!***************************************************************

IMPLICIT none


CONTAINS


SUBROUTINE get_unit ( iunit )

!*************************************************************
!
! GET_UNIT returns a free FORTRAN unit number.
!
!  Discussion:
!
!    A "free" FORTRAN unit number is an integer between 1 and 99 which
!    is not currently associated with an I/O device.  A free FORTRAN unit
!    number is needed in order to open a file with the OPEN command.
!
!    If IUNIT = 0, then no free FORTRAN unit could be found, although
!    all 99 units were checked (except for units 5, 6 and 9, which
!    are commonly reserved for console I/O).
!
!    Otherwise, IUNIT is an integer between 1 and 99, representing a
!    free FORTRAN unit.  Note that GET_UNIT assumes that units 5 and 6
!    are special, and will never return those values.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    26 October 2008
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Output, integer ( kind = 4 ) IUNIT, the free unit number.
!
!*************************************************************

IMPLICIT none


! local variables
integer ( kind = 4 ) i
integer ( kind = 4 ) ios
integer ( kind = 4 ) iunit
logical              lopen


!*************************************************************
!	begin code
!*************************************************************
iunit = 0
do i = 1, 99

	if ( i /= 5 .and. i /= 6 .and. i /= 9 ) then

		inquire ( unit = i, opened = lopen, iostat = ios )

		if ( ios == 0 ) then

			if ( .not. lopen ) then

				iunit = i
				return
			end if
		end if
	end if
end do


END SUBROUTINE get_unit


SUBROUTINE csv_file_read( file_name, values, mm, nn, istat )


!*****************************************************************************
!
!	CSV_FILE_READ reads data from a CSV file
!
!  Modified:
!
!    28 AUGUST 2012
!
!  Author:
!
!    Justin van de Ven
!
!  Parameters:
!
!	file_name	: full name of file to be saved
!	values		: matrix to read data into
!	mm			: number of rows in matrix
!	nn			: number of cols in matrix
!	istat		: error flag
!
!*****************************************************************************

IMPLICIT NONE


integer(4), intent(in) :: mm, nn
integer(4), intent(out) :: istat
real(8), intent(out) :: values(mm,nn)
character(*) :: file_name

! local variables
integer(4) :: istat1, file_id, ii


!*****************************************************************************
!	begin code
!*****************************************************************************
values = 0.0

! open file to read from
istat = 0
call get_unit( file_id )
open ( unit = file_id, file = file_name, action = 'read', iostat = istat1 )
if ( istat1.ne.0 ) then

	istat = istat1
else
! read data from file

	do ii = 1, mm

		if ( istat.eq.0 ) then

			read ( file_id, *, IOSTAT=istat1 ) values(ii,:)
			istat = istat1
		end if
	end do
	close ( file_id )
end if


END SUBROUTINE csv_file_read

	
SUBROUTINE csv_file_write( file_name, values, mm, nn, flag, sig_figs )

!*************************************************************
!
!	CSV_FILE_WRITE writes real data to CSV file
!
!  Modified:
!
!    28 APRIL 2015
!
!  Author:
!
!    Justin van de Ven
!
!  Parameters:
!
!	file_name		: full name of file to be saved
!	values			: values to be written
!	mm				: number of rows in matrix
!	nn				: number of cols in matrix
!	flag			: 0 if unassigned, 1 if integer, 2 if currency, 3 if other real
!	sig_figs		: significant figures to save for (other) real variable (OPTIONAL)
!
!*************************************************************

IMPLICIT none


integer(4), intent(in) :: mm, nn, flag
integer(4), intent(in), optional :: sig_figs
real(8), intent(in) :: values(mm, nn)
character(*) :: file_name

! local variables
integer(4) :: file_id, istat, ii, jj
character(nn*24) :: output_line


!*************************************************************
!	begin code
!*************************************************************


!*************************************************************
!	open file to write to
!*************************************************************
open ( newunit = file_id, file = file_name, status = 'replace', action = 'write', iostat = istat )
if ( istat.ne.0 ) then

	if ( istat.ne.0 ) then

		print *, 'error writing CSV file'
		print *, 'could not open "' // trim ( file_name ) // '".'
		pause 'press enter to exit'
		continue
		stop
		continue
	end if
end if


!*************************************************************
!	write output to file
!*************************************************************
do ii = 1, mm

	output_line = '  '
	do jj = 1, nn

		call csv_data_append( flag, values(ii,jj), output_line, sig_figs )
	end do
	write ( file_id, '(a)' ) trim ( output_line )
end do
close ( file_id, status = 'save', iostat = istat )
if ( istat.ne.0 ) then

	continue
end if


END SUBROUTINE csv_file_write


SUBROUTINE csv_data_append( flag, datum, output_line, sig_figs )

!*************************************************************
!
! CSV_DATA_APPEND appends new variable onto the end of a CSV record.
!
!  Parameters:
!
!	flag:		0 = unassigned, 1 = integer, 2 = currency, 3 = real
!	datum:		number to be added to text file
!	output line: the CSV record
!	sig_figs:	number of significant figures to be written (for real numbers)
!
!*************************************************************

IMPLICIT none


integer(4), intent(in), optional :: sig_figs
integer(4), intent(in) :: flag
real(8), intent(in) :: datum
character(*) :: output_line

! local variables
integer(4) :: ii, intgr, numb_width, flag_h, sig_figs1
character(20) :: fmat


!*************************************************************
!	begin code
!*************************************************************
! locate last used location in RECORD.
ii = len_trim ( output_line )
if ( present(sig_figs) ) then

	sig_figs1 = min(15,sig_figs)
else

	sig_figs1 = 5
end if

! append comma
if ( ii.gt.0 ) then
! not first record

	ii = ii + 1
	output_line(ii:ii) = ','
end if

! append value after identifying relevant detail
if ( datum.eq.0.0 ) then
! record as 0

	ii = ii + 1
	output_line(ii:ii) = '0'
else

	if ( flag.eq.0 ) then

		if ( datum.eq.real(idint(datum)) ) then
		! record as integer

			flag_h = 1
		else

			flag_h = 3
		end if
	else

		flag_h = flag
	end if
	if ( flag_h.eq.1 ) then
	! treat as integer

		intgr = nint( datum )
		call i4_width( numb_width, intgr )
		write ( fmat, '(a,i2,a)' ) '(i', numb_width, ')'
		write ( output_line(ii+1:ii+numb_width), fmat ) intgr
	elseif ( flag_h.eq.2 ) then
	! treat as currency

		call c_width( numb_width, datum )
		write ( fmat, '(a,i2,a)' ) '(f', numb_width, '.2)'
		write ( output_line(ii+1:ii+numb_width), fmat ) datum
	else
	! treat as general real

		if ( sig_figs1.le.5 ) then

			call r8_stat( fmat, flag_h, sig_figs1, numb_width, datum )
			if ( flag_h.eq.1 ) then
			! treat as an integer

				intgr = nint( datum )
				write ( output_line(ii+1:ii+numb_width), fmat ) intgr
			else

				write( output_line(ii+1:ii+numb_width), fmat ) datum
			end if
		else

			numb_width = 4 + sig_figs1 + 3
			write ( fmat, '(a,i2,a,i2,a)' ) '(1PG', numb_width,'.', sig_figs1,'E3)'
			write( output_line(ii+1:ii+numb_width), fmat ) datum
		end if
	end if
end if


END SUBROUTINE csv_data_append


SUBROUTINE i4_width( numb_width, ii )

!*************************************************************
!
! I4_WIDTH returns the "width" of an I4.
!
!  Discussion:
!
!    The width of an integer is the number of characters necessary to print it.
!
!    The width of an integer can be useful when setting the appropriate output
!    format for a vector or array of values.
!
!    An I4 is an integer ( kind = 4 ) value.
!
!  Example:
!
!		II  I4_WIDTH
!    -----  -------
!    -1234    5
!     -123    4
!      -12    3
!       -1    2
!        0    1
!        1    1
!       12    2
!      123    3
!     1234    4
!    12345    5
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    30 April 2015
!
!  Author:
!
!    John Burkardt
!	 Justin van de Ven (modified)
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) II, the number whose width is desired.
!
!    Output, integer ( kind = 4 ) numb_width, the number of characters
!    necessary to represent the integer in base 10, including a negative
!    sign if necessary.
!
!*************************************************************

IMPLICIT none


integer(4), intent(in) :: ii
integer(4), intent(out) :: numb_width

! local variables
integer(4) :: ii_abs, kk


!*************************************************************
!	begin code
!*************************************************************
if ( ii.eq.0 ) then

	numb_width = 1
else

	ii_abs = abs(ii)
	numb_width = 0
	kk = 1
	do while ( ii_abs.ge.kk )

		numb_width = numb_width + 1
		kk = kk * 10
	end do
	if ( ii.lt.0 ) then

		numb_width = numb_width + 1
	end if
end if


END SUBROUTINE i4_width


SUBROUTINE c_width( numb_width, rr )

!*************************************************************
!
!  C_WIDTH returns the "width" of a currency variable rr
!
!  date:
!    30 April 2015
!
!  Author:
!	 Justin van de Ven
!
!  Parameters:
!
!    Input, real(8) rr, the number whose width is desired.
!
!    Output, integer(4) numb_width, the number of characters
!    necessary to represent rr in base 10, including a negative
!    sign if necessary.
!
!*************************************************************

IMPLICIT none


real(8), intent(in) :: rr
integer(4), intent(out) :: numb_width

! local variables
integer(4) :: rr_int


!*************************************************************
!	begin code
!*************************************************************
rr_int = nint(rr*100.0)
call i4_width( numb_width, rr_int )
numb_width = max(4, numb_width+1)


END SUBROUTINE c_width


SUBROUTINE r8_stat( fmat, flag_h, sig_figs1, numb_width, rr )

!*************************************************************
!
!  r8_stat returns the format (fmat), and "width" (numb_width)
!  of a real variable (rr), up to significant figures (sig_figs)
!
!  date:
!    30 April 2015
!
!  Author:
!	 Justin van de Ven
!
!  Parameters:
!
!    Input, real(8) rr, the number whose width is desired.
!
!    Output, character(integer(4) numb_width, the number of characters
!    necessary to represent rr in base 10, including a negative
!    sign if necessary.
!
!*************************************************************

IMPLICIT none


integer(4) :: flag_h
integer(4), intent(in) :: sig_figs1
real(8), intent(in) :: rr
integer(4), intent(out) :: numb_width
character(20), intent(out) :: fmat

! local variables
integer(4) :: base10, sig_figs_h, dec_width, flag_fp, rr_int
real(8) :: rr_abs0, rr_abs1, rr_error


!*************************************************************
!	begin code
!*************************************************************
flag_h = 0
sig_figs_h = sig_figs1
rr_abs1 = abs(rr)
base10 = int(log(rr_abs1) / log(10.0))

if ( base10.lt.0 ) then

	rr_abs0 = rr_abs1 * (10.0 ** real(sig_figs_h-base10))
else

	rr_abs0 = rr_abs1 * (10.0 ** real(sig_figs_h-1-base10))
end if
rr_abs1 = real(nint(rr_abs0 / 10.0)) * 10.0
rr_error = rr_abs1 - rr_abs0
do while ( nint(rr_error).eq.0 )

	sig_figs_h = sig_figs_h - 1
	rr_abs0 = rr_abs1 / 10.0
	rr_abs1 = real(nint(rr_abs0 / 10.0)) * 10.0
	rr_error = rr_abs1 - rr_abs0
end do


!*************************************************************
!	determine whether Exponential or Fixed point decimal more efficient
!*************************************************************
if ( sig_figs_h.eq.1 ) then

	if ( (base10.lt.6).and.(base10.gt.-4) ) then
	! fixed point format

		flag_fp = 1
	else
	! exponential format

		flag_fp = 0
	end if
else

	if ( (base10.lt.(sig_figs_h+4)).and.(base10.gt.-3) ) then
	! fixed point format

		flag_fp = 1
	else
	! exponential format

		flag_fp = 0
	end if
end if


!*************************************************************
!	evaluate width and fmat identifiers
!*************************************************************
if ( flag_fp.eq.1 ) then
! fixed format

	if ( base10.ge.(sig_figs_h-1) ) then
	! assign as integer

		flag_h = 1
		rr_int = nint(rr)
		call i4_width( numb_width, rr_int )
		write ( fmat, '(a,i2,a)' ) '(i', numb_width, ')'
	else
	! assign as a decimal

		dec_width = sig_figs_h - base10
		if ( abs(rr).gt.1 ) then

			dec_width = max(1, dec_width - 1)
		end if

		numb_width = max(0, base10) + 2 + dec_width
		if ( rr.lt.0.0 ) then

			numb_width = numb_width + 1
		end if
		write ( fmat, '(a,i2,a,i2,a)' ) '(f', numb_width, '.', dec_width, ')'
	end if
else
! exponential

	dec_width = 1 + max(0, sig_figs_h - 2)
	numb_width = 6 + dec_width
	if ( rr.lt.0.0 ) then

		numb_width = numb_width + 1
	end if
	write ( fmat, '(a,i2,a,i2,a)' ) '(ES', numb_width, '.', dec_width, ')'
end if


END SUBROUTINE r8_stat


END MODULE Comm_tools	