MODULE solve_continuous


!*************************************************************
!
!	Module to manage solution of utility maximisation problem
!	over all continuous decision variables
!
!*************************************************************

IMPLICIT none


CONTAINS


SUBROUTINE continuous_choice(xx, ndim_inner, n_iuser, n_ruser, training_state, leisure, labour, prwage1, prwage2, emp1, emp2, nwos1, training_choice, training_gain, pen_regime, pens_takeup, ri_am1, pc_am1, pcemp_am1, isa_am1, u, val, dv, dv2)

!*************************************************************
!
!	Routine to solve for utility maximising portfolio allocation
!
!	This routine is fed in the current state location, a set
!	a set of decisions on discrete space alternatives, and
!	initial guesses for continuous space control variables.
!	The routine then identifies the feasible search space for
!	the continuous variables, and optimises over this space.
!
!*************************************************************

USE global
USE global_ParamStore
USE global_CommonAll
USE global_CommonCohort
USE global_f1
USE analysis_tools
USE optimisation
USE val_fn_evals

IMPLICIT none


integer(4) :: ndim_inner, n_iuser, n_ruser, training_choice, pens_takeup, emp1, emp2, nwos1, pen_regime
real(8) :: xx(ndim_inner), training_state, leisure, labour, prwage1, prwage2, training_gain
real(8) :: ri_am1, pc_am1, isa_am1, u, val, dv, dv2, pcemp_am1

! local variables
integer(4) :: ii, nn, invest_id, pc_id, isa_id, istat, n_exp
integer(4) :: iuser(n_iuser)
real(8) :: ri_a, ri_cc, pc_a, pc_cc, isa_a, isa_cc, bl_c, bu_c
real(8) :: ruser(n_ruser), lab_inc
real(8) :: bl(10), bu(10), gg(10)


!****************************************************************
!	begin code
!****************************************************************


!****************************************************************
!	determine dimensionality of decision space
!****************************************************************
if ( (model_par(47).gt.0.5).and.(current_age.lt.nint(model_par(7))) ) then
! need to consider isa investments

	isa_id = 1
else

	isa_id = 0
end if
if ( (xx(1).gt.max(10.0,macro_par(15))).and.(model_par(33).gt.0.5) ) then
! need to solve portfolio allocation

	invest_id = 1
else

	invest_id = 0
end if
if ( (pen_par(11).gt.0.5).and.(pen_par(13).gt.0.5).and.((emp1+emp2).gt.0).and.(pen_regime.gt.0) ) then
! endogenous PP contributions

	pc_id = 1
else

	pc_id = 0
end if


!****************************************
! initialise optimisation brackets
!****************************************
if ( invest_id.eq.1 ) then
! optimise over safe and risky assets

	ri_a = 0.0_8
	ri_cc = 1.0_8
else

	ri_a = 0.0_8
	ri_cc = 0.0_8
	ri_am1 = 0.0_8
end if
if ( pc_id.eq.1 ) then
! optimise over pension contribution rates

	if ( semp_state.eq.0 ) then

		if ( pen_regime.eq.1 ) then

			if ( pen_par(14).gt.0.5 ) then

				pc_a = pen_par(51+pcr_state)
			else

				pc_a = pen_par(27)
			end if
		elseif ( pen_regime.eq.2 ) then

			if ( pen_par(14).gt.0.5 ) then

				pc_a = pen_par(144+pcr_state)
			else

				pc_a = pen_par(145)
			end if
		else

			!$OMP CRITICAL(error_1)
			print *, 'problem in evaluating limits for private pension contributions'
			pause 'press enter to exit'
			stop
			continue
			!$OMP END CRITICAL(error_1)
		end if
	else

		pc_a = semp_par(24)
	end if
	pc_cc = pen_par(28)
else

	pc_a = pc_am1
	pc_cc = pc_am1
end if
if ( isa_id.eq.1 ) then
! optimise over isa contributions

	isa_a = -xx(isa_indx_am1)*macro_par(26)
	isa_cc = macro_par(25) * model_index(current_year-year0_param+1,11) * na_g / 52.0
else

	if ( model_par(47).gt.0.5 ) then
		
		isa_a = -xx(isa_indx_am1)*macro_par(26)
		isa_cc = isa_a
		isa_am1 = isa_a
	else
		
		isa_a = 0.0
		isa_cc = 0.0
		isa_am1 = 0.0
	end if
end if


!*********************************************
!	pack optimisation user vectors
!*********************************************
! initialise
bl = 0.0
bu = 0.0
gg = 0.0
iuser = 0
ruser = 0.0
istat = 0

! set values
ii = 1
if ( pc_id.eq.1 ) then

	ii = ii + 1
	bl(ii) = pc_a
	bu(ii) = pc_cc
	gg(ii) = pc_am1
else

	ruser(51) = pc_am1
end if
if ( isa_id.eq.1 ) then

	ii = ii + 1
	bl(ii) = isa_a
	bu(ii) = isa_cc
	gg(ii) = isa_am1
else

	ruser(52) = isa_am1
end if
if ( invest_id.eq.1 ) then

	ii = ii + 1
	bl(ii) = ri_a
	bu(ii) = ri_cc
	gg(ii) = ri_am1
else

	ruser(53) = ri_am1
end if
nn = ii

iuser(1) = 1		! function flag
iuser(11) = pc_id
iuser(12) = isa_id
iuser(13) = invest_id

iuser(21) = ndim_inner
iuser(22) = emp1
iuser(23) = emp2
iuser(24) = training_choice
iuser(25) = pen_regime
iuser(26) = pens_takeup
iuser(27) = nwos1

ruser(11:10+ndim_inner) = xx
ruser(31) = pcemp_am1
ruser(32) = training_state
ruser(33) = leisure
ruser(34) = labour
ruser(35) = prwage1
ruser(36) = prwage2
ruser(38) = training_gain


!*****************************************
! re-check expectations vectors
!*****************************************
call expectations_update(nn, gg(1:nn), bl_c, bu_c, n_iuser, iuser, n_ruser, ruser, 1, istat)
if ( istat.eq.1 ) then
! consumption choice here from a zero set

	ii = 1
	if ( pc_id.eq.1 ) then

		ii = ii + 1
		if ( (gg(ii)-bl(ii)).gt.1.0E-5 ) then

			istat = 0
			gg(ii) = bl(ii)
		end if
	end if
	if ( isa_id.eq.1 ) then

		ii = ii + 1
		if ( (gg(ii)-bl(ii)).gt.1.0E-5 ) then

			istat = 0
			gg(ii) = bl(ii)
		end if
	end if
	if ( invest_id.eq.1 ) then

		ii = ii + 1
		if ( (gg(ii)-bl(ii)).gt.1.0E-5 ) then

			istat = 0
			gg(ii) = bl(ii)
		end if
	end if
	if ( istat.eq.0 ) then
	! check if new starting values work

		call expectations_update(nn, gg(1:nn), bl_c, bu_c, n_iuser, iuser, n_ruser, ruser, 1, istat)
	end if
end if
if ( (pen_par(11).gt.0.5).and.(current_age.le.nint(model_par(5))).and.(istat.eq.0) ) then
! verify that labour incomes permit pension contributions

	lab_inc = max(ruser(108), ruser(109))
	if ( ( ((prec_state+pens_takeup).eq.0).or.(tax_par4(59).gt.0.5) ).and.(pen_par(11).gt.0.5).and.(pen_regime.ne.0).and. &
		( (semp_state.eq.0).or.(semp_par(2).gt.0.5) ).and. &
		! private pension contributions possible
		( lab_inc.gt.(pen_par(32)*(model_index(current_year-year0_param+1,10))) ).and. &
		( lab_inc.lt.(pen_par(33)*(model_index(current_year-year0_param+1,10))) ) ) then

		continue
	else

		pc_am1 = 0.0_8
		pcemp_am1 = 0.0_8
		if ( pc_id.eq.1 ) then
		! need to reset optimisation flags and vectors

			pc_id = 0
			iuser(11) = pc_id
			ii = 1
			ruser(51) = pc_am1
			if ( isa_id.eq.1 ) then

				ii = ii + 1
				bl(ii) = isa_a
				bu(ii) = isa_cc
				gg(ii) = isa_am1
			end if
			if ( invest_id.eq.1 ) then

				ii = ii + 1
				bl(ii) = ri_a
				bu(ii) = ri_cc
				gg(ii) = ri_am1
			end if
			bl(ii+1) = 0.0
			bu(ii+1) = 0.0
			gg(ii+1) = 0.0
			nn = ii
		end if
	end if
end if


!*********************************************
!	pass to search routines
!*********************************************
if ( istat.eq.2 ) then
! this option not feasible

	pc_id = 0
	isa_id = 0
	invest_id = 0
	gg(1) = 0.0
	val = 999.0
elseif ( istat.eq.1 ) then
! consumption choice from a zero set - adopt constrained solution

	gg(1) = macro_par(11)
	n_exp = iuser(201)
	call val_eval(n_exp, gg(1), n_iuser, iuser, n_ruser, ruser, val)
	val = - val
elseif ( model_par(16).gt.0.5 ) then
! consumption exogenously set
	
	if ( bu_c.lt.u ) then
		
		u = bu_c
	end if
	gg(1) = u
	n_exp = iuser(201)
	call val_eval(n_exp, gg(1), n_iuser, iuser, n_ruser, ruser, val)
	val = - val
else

	ii = 1
	bl(ii) = bl_c
	bu(ii) = bu_c
	gg(ii) = (bl(ii) + bu(ii))/2.0_8
	ruser(91:100) = bl
	call fn_minimise( nn, bl(1:nn), bu(1:nn), gg(1:nn), val, 0, n_iuser, iuser, n_ruser, ruser, istat )
end if


!*********************************************
!	unpack user vectors
!*********************************************
ii = 1
u = gg(ii)
if ( pc_id.eq.1 ) then

	ii = ii + 1
	pc_am1 = gg(ii)
end if
if ( isa_id.eq.1 ) then

	ii = ii + 1
	isa_am1 = gg(ii)
end if
if ( invest_id.eq.1 ) then

	ii = ii + 1
	ri_am1 = gg(ii)
end if
val = -val
dv = ruser(401)
dv2 = ruser(402)


END SUBROUTINE continuous_choice


SUBROUTINE define_iuser_util(n_iuser, iuser)

!*************************************************************
!
!	This is a dummy Routine that keeps track of the definition of
!	iuser by utility optimsation routines
!
!*************************************************************

IMPLICIT none


integer(4), intent(in) :: n_iuser
integer(4) :: iuser(n_iuser)


!*************************************************************
!	begin code
!
! iuser organised as follows:
! 	1  	: flag indicates the fucntion to be optimised 1 = utility maximisation, 2 = banana, 3 = ...
!	11-20	: cells allocated for flags indicating alternative decision dimensions (up to 10)
!	21-100	: cells allocated for exogenous input prior to any function optimisations
!	101-200	: cells allocated for variables required to update expectations
!		101-110 : ZONE 1
!		111-115 : ZONE 2
!		116-120 : ZONE 3
!	201-300+nexp_d1	: cells allocated for input required to evaluate value function
!
!*************************************************************

iuser = 0
!iuser(1) = 1		! value of 1 indicates that we are solving utility maximisation decisions
!iuser(11) = pc_id
!iuser(12) = isa_id
!iuser(13) = invest_id

!iuser(21) = ndim_inner
!iuser(22) = emp1
!iuser(23) = emp2
!iuser(24) = training_choice
!iuser(25) = pen_regime
!iuser(26) = pens_takeup
!iuser(27) = nwos1

! ZONE 1
! iuser(101) = ret_annuity
! iuser(102) = pen_receive
! iuser(103) = edu_trans
! iuser(104) = nexp_here0

! ZONE 2
! iuser(111) = isa_uncert
! iuser(112) = nexp_here0

! ZONE 3
! iuser(116) = cont_out
! iuser(117) = nexp_here0
! iuser(118) = random_here
! iuser(119) = dppart0

! iuser(199) = nexp_here0

! OUTPUTS FOR UTILITY MAXIMISATIONS
! iuser(201) = n_exp
! iuser(301:300+nexp_d1) = nexp_c


END SUBROUTINE define_iuser_util


SUBROUTINE define_ruser_util(n_ruser)

!*************************************************************
!
!	This is a dummy Routine that keeps track of the definition of
!	ruser by utility optimsation routines
!
!*************************************************************

IMPLICIT none


integer(4), intent(in) :: n_ruser
real(8) :: ruser(n_ruser)


!*************************************************************
!	begin code
!
! ruser organised as follows:
!	1-10	: cells allocated to indicate decision vector of last expectations update
!	11-30	: cells allocated to record continuous state variables of current call
!	31-90	: cells allocated for exogenous input prior to any function optimisations
!	91-100	: cells allocated to record lower bounds on elements of decision vector
!	101-200	: cells allocated for variables required to update expectations
!		101:130 : ZONE 1 + 501:750 for exp0 vectors
!		131:150	: ZONE 2 + 751:1000 for exp0 vectors
!		151:170 : ZONE 3 + 1001:1250 for exp0 vectors
!	201-300	: cells allocated for input required to evaluate value function
!	401-500 : value function output
!
!*************************************************************
ruser = 0.0
!ruser(1:nn) = gg(1:nn)

!ruser(11:10+ndim_inner) = xx

!ruser(31) = pcemp_am1
!ruser(32) = training_state
!ruser(33) = leisure
!ruser(34) = labour
!ruser(35) = prwage1
!ruser(36) = prwage2
!ruser(38) = training_gain

!ruser(51) = pc_am1	! if pc_am1 not optimised over the intensive margin
!ruser(52) = isa_am1	! if isa_am1 not optimised over the intensive margin
!ruser(53) = ri_am1	! if ri_am1 not optimised over the intensive margin

!ruser(91:90+ndim_inner) = bl

! ZONE 1
! ruser(101) = uipp_contbn
! ruser(102) = max_oph
! ruser(103) = max_pph
! ruser(104) = training_cost
! ruser(105) = train_experience
! ruser(106) = pr_emp
! ruser(107) = lab_inc
! ruser(108) = lab_inc1
! ruser(109) = lab_inc2
! ruser(110) = lab_inc1f
! ruser(111) = ppen_inc
! ruser(112) = spen_inc
! ruser(117) = vat0_0
! ruser(118) = vat0_1
! ruser(119) = vat1_0

! ruser(501+nexp_g0*0:500+nexp_g0*1) = prob_exp0
! ruser(501+nexp_g0*1:500+nexp_g0*2) = w_exp0
! ruser(501+nexp_g0*2:500+nexp_g0*3) = op_exp0
! ruser(501+nexp_g0*3:500+nexp_g0*4) = pp_exp0
! ruser(501+nexp_g0*4:500+nexp_g0*5) = ob_exp0
! ruser(501+nexp_g0*5:500+nexp_g0*6) = cp1_exp0
! ruser(501+nexp_g0*6:500+nexp_g0*7) = cp2_exp0
! ruser(501+nexp_g0*7:500+nexp_g0*8) = isa_exp0
! ruser(501+nexp_g0*8:500+nexp_g0*9) = ui_exp0

! ZONE 2
! ruser(131) = isa_hh
! ruser(132) = isa_cost
! ruser(136:135+nexp_g0) = isa_inc

! ruser(751+nexp_g0*0:750+nexp_g0*1) = prob_exp0
! ruser(751+nexp_g0*1:750+nexp_g0*2) = w_exp0
! ruser(751+nexp_g0*2:750+nexp_g0*3) = op_exp0
! ruser(751+nexp_g0*3:750+nexp_g0*4) = pp_exp0
! ruser(751+nexp_g0*4:750+nexp_g0*5) = ob_exp0
! ruser(751+nexp_g0*5:750+nexp_g0*6) = cp1_exp0
! ruser(751+nexp_g0*6:750+nexp_g0*7) = cp2_exp0
! ruser(751+nexp_g0*7:750+nexp_g0*8) = isa_exp0
! ruser(751+nexp_g0*8:750+nexp_g0*9) = ui_exp0

! ZONE 3
! ruser(153) = OP_contbn
! ruser(154) = PP_contbn
! ruser(156) = pen_cost
! ruser(158) = op_priv_penc
! ruser(159) = pp_priv_penc

! ruser(1001+nexp_g0*0:1000+nexp_g0*1) = prob_exp0
! ruser(1001+nexp_g0*1:1000+nexp_g0*2) = w_exp0
! ruser(1001+nexp_g0*2:1000+nexp_g0*3) = op_exp0
! ruser(1001+nexp_g0*3:1000+nexp_g0*4) = pp_exp0
! ruser(1001+nexp_g0*4:1000+nexp_g0*5) = ob_exp0
! ruser(1001+nexp_g0*5:1000+nexp_g0*6) = cp1_exp0
! ruser(1001+nexp_g0*6:1000+nexp_g0*7) = cp2_exp0
! ruser(1001+nexp_g0*7:1000+nexp_g0*8) = isa_exp0
! ruser(1001+nexp_g0*8:1000+nexp_g0*9) = ui_exp0

! OUTPUT FOR UTILITY MAXIMISATIONS
! start_indx_here = 1500
! ruser(start_indx_here+1+max_prob*0:start_indx_here+max_prob*1) = prob_exp
! ruser(start_indx_here+1+max_prob*1:start_indx_here+max_prob*2) = y_exp
! ruser(start_indx_here+1+max_prob*2:start_indx_here+max_prob*3) = cost_exp
! ruser(start_indx_here+1+max_prob*3:start_indx_here+max_prob*4) = vat0
! ruser(start_indx_here+1+max_prob*4:start_indx_here+max_prob*5) = vat1
! do ii = 1, ndim_a

! 	ruser(start_indx_here+1+max_prob*(4+ii):start_indx_here+max_prob*(5+ii)) = x_exp(:,ii)
! end do


END SUBROUTINE define_ruser_util


END MODULE solve_continuous
